from __future__ import annotations

from dataclasses import dataclass
from typing import Callable, Generator, Optional, Union

feedback_generator = Generator[Union[str, "feedback_generator"], None, None]
command_callback = Callable[[str], feedback_generator]


@dataclass
class Keyword:
    synonyms: tuple[str]
    weight: float


@dataclass
class Command:
    identifier: str
    keywords: tuple[Keyword]
    callback: command_callback



from ..config import INPUT_DEVICE_CONFIG, VAD_CONFIG, get_config
from .audio_api import api
from .audio_processing_utils import mean_spectogram_from_buffer


class VADetector:
    def __init__(self, frame_count=512, pause_threshold=1) -> None:
        self.frame_count = frame_count
        self.pause_threshold = pause_threshold

        self.__input_device_config = get_config(INPUT_DEVICE_CONFIG)
        self.__vad_config = get_config(VAD_CONFIG)

    def listen(self):
        segment = b""
        started_talking = False
        pause_time = 0
        detection_threshold = self.__vad_config["min_speech_volume_ratio"] * (
            self.__vad_config["speech_level"] - self.__vad_config["noise_level"]
        )
        with api.open(**self.__input_device_config) as stream:
            while stream.is_active():
                buffer = stream.read(self.frame_count, exception_on_overflow=False)
                volume = mean_spectogram_from_buffer(buffer)

                if volume >= detection_threshold:
                    started_talking = True
                    pause_time = 0
                elif started_talking:
                    pause_time += (
                        1 / self.__input_device_config["rate"]
                    ) * self.frame_count

                if started_talking:
                    segment += buffer
                    if pause_time >= self.pause_threshold:
                        yield segment
                        # reset values
                        segment = b""
                        started_talking = False
                        pause_time = 0



from pyaudio import paInt16

from .config import make_config
from .core.executor import execute_command
from .core.interpretation import CommandInterpreter
from .core.recognition import Model
from .core.voice_activity_detection import VADetector
from .types import Command


class Assistant:
    def __init__(
        self,
        wake_word: str,
        commands: list[Command],
        model_name: str,
        input_device_query: str,
        min_speech_volume_ratio: float = 0.2,
        input_device_dtype: int = paInt16,
        verbose=False,
    ) -> None:
        make_config(input_device_query, min_speech_volume_ratio, input_device_dtype)
        self.commands = commands
        self.verbose = verbose

        self.vad = VADetector()
        self.model = Model(model_name)
        self.interpreter = CommandInterpreter(self.commands, wake_word)
        self.__segment_generator = self.vad.listen()

        self._awake = False

    @property
    def is_awake(self):
        return self._awake

    def next_transcription(self):
        segment = next(self.__segment_generator)
        return self.model.transcribe(segment)

    def run(self):
        for segment in self.__segment_generator:
            self.on_voice_activity_detection(segment)
            transcription = self.model.transcribe(segment)
            self.on_transcription(transcription)
            if self._awake:
                if command := self.interpreter.associate(transcription):
                    self.on_command(command)
                    execute_command(command, transcription, self)
                    self._awake = False
                else:
                    self.on_no_associated_command()
            else:
                if self.interpreter.is_wake_word(transcription):
                    self.on_awake()
                    self._awake = True

    def on_voice_activity_detection(self, segment: bytes):
        if self.verbose:
            print("Detected Voice.")

    def on_transcription(self, transcription: str):
        if self.verbose:
            print(f"Transcribed: {transcription}")

    def on_command(self, command: Command):
        if self.verbose:
            print(f"Executing Command: {command.identifier}")

    def on_no_associated_command(self):
        if self.verbose:
            print(f"Couldn't associate transcription with command.")

    def on_awake(self):
        if self.verbose:
            print("Assistant just woke up.")



import zipfile
from io import BytesIO
from pathlib import Path

import requests
from tqdm import tqdm

ROOT = Path("~/.cache/voice_commands/models").expanduser()


def download_model(
    url: str,
    name: str,
    root: str = ROOT,
):
    root = Path(root).expanduser()
    (root / name).mkdir(parents=True, exist_ok=False)

    response = requests.get(url, stream=True)
    content_size = int(response.headers.get("content-length", 0))
    block_size = 1024

    data = BytesIO()

    with tqdm(
        total=content_size,
        unit="iB",
        unit_scale=True,
        desc=f"Downloading <{name}> from <{url}>",
    ) as pbar:
        for chunk in response.iter_content(chunk_size=block_size):
            data.write(chunk)
            pbar.update(len(chunk))

    with zipfile.ZipFile(data, "r") as file:
        with tqdm(
            total=len(data.getvalue()),
            unit="iB",
            unit_scale=True,
            desc="Extracting",
        ) as pbar:
            for file_info in file.infolist():
                file.extract(file_info, root / name)
                pbar.set_postfix(file=file_info.filename)
                pbar.update(file_info.file_size)


def list_models():
    for model_name_dir in ROOT.iterdir():
        print(model_name_dir.name)



"""
This Module is a representation of my hatred towards pyaudio
its just context managers
"""

from pyaudio import *


class Api(PyAudio):
    def open(self, *args, **kwargs):
        stream = AudioStream(self, *args, **kwargs)
        self._streams.add(stream)
        return stream

    @staticmethod
    def get_input_devices():
        devices = []

        for i in range(api.get_device_count()):
            device_info = api.get_device_info_by_index(i)
            if device_info["maxInputChannels"] > 0:
                devices.append({"index": i, "name": device_info["name"]})

        return devices

    @staticmethod
    def get_output_devices():
        devices = []

        for i in range(api.get_device_count()):
            device_info = api.get_device_info_by_index(i)
            if device_info["maxOutputChannels"] > 0:
                devices.append({"index": i, "name": device_info["name"]})

        return devices

    @staticmethod
    def get_input_device_from_query(query: str):
        devices = Api.get_input_devices()

        for device in devices:
            if query.lower() in device["name"].lower():
                return device


class AudioStream(Stream):
    def __enter__(self):
        self.start_stream()
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.close()


api = Api()



import math
import string

from ..types import Command


class CommandInterpreter:
    """This implements the Interpreting-Procedure. Subclass this and override its methods to change its behaviour."""

    def __init__(
        self, commands: list[Command], wake_word: str, similarity_threshold: int = 0.8
    ) -> None:
        self.commands = commands
        self.wake_word = self.format_text(wake_word)
        self.similarity_threshold = similarity_threshold

    @staticmethod
    def format_text(text) -> str:
        """Returns lowercase `text` without punctuation

        Args:
            text (_type_): the text

        Returns:
            str: `text` without punctuation and lowercase
        """
        return text.translate(str.maketrans("", "", string.punctuation)).lower()

    @staticmethod
    def get_similarity(str1: str, str2: str) -> float:
        """Calculates the jaccard-simularity of two strings

        Args:
            str1 (_type_): str1
            str2 (_type_): str2

        Returns:
            float: similarity in percent
        """
        set1 = set(str1.split())
        set2 = set(str2.split())
        intersection_size = len(set1.intersection(set2))
        union_size = len(set1.union(set2))

        return intersection_size / union_size if union_size != 0 else 0

    def calculate_word_score(self, word: str, text: str) -> float:
        """Gets the similarity for each word from text and `word` and returns the highscore

        Args:
            word (str): the word
            text (str): the transcription

        Returns:
            float: highscore
        """
        highscore = 0
        for word2 in text.split():
            similarity = self.get_similarity(word, word2)
            if similarity > highscore:
                highscore = similarity

        return highscore

    def calculate_command_score(self, command: Command, text: str) -> float:
        """Calculates the score of a command given the transcribed text by
        taking into account the weight of each keyword (synonyms are
        treated as options but they are ordered hierarchically) and its
        length. The highest scoring synonyms score + sqrt of its length
        adds to score if it is greater then similarity_threshold though

        Args:
            command (Command): the command
            text (str): the transcribed text

        Returns:
            float: the commands score as element of [0; oo[
        """
        score = 0
        for keyword in command.keywords:
            synonym_highscore = 0
            leading_synonym = None
            for synonym in keyword.synonyms:
                synonym_score = self.calculate_word_score(synonym, text)
                if synonym_score > synonym_highscore:
                    synonym_highscore = synonym_score
                    leading_synonym = synonym

            if synonym_highscore >= self.similarity_threshold:
                score += synonym_score + math.sqrt(len(leading_synonym))

        return score

    def associate(self, transcription: str, score_threshold: float = 0) -> Command:
        """Gets the command from `self.command_list` that has the highest score

        Args:
            transcription (str): the transcription
            score_threshold (float, optional): Minimum score that a command has to reach to be counted.
                                               If no command reaches the threshold `None` is returned. Defaults to 0.

        Returns:
            Command: _description_
        """
        text = self.format_text(transcription)
        selected_command = None
        highscore = 0
        for command in self.commands:
            score = self.calculate_command_score(command, text)
            if score > highscore and score >= score_threshold:
                selected_command = command

        return selected_command

    def is_wake_word(self, transcription: str):
        if (
            self.calculate_word_score(self.wake_word, self.format_text(transcription))
            >= self.similarity_threshold
        ):
            return True
        return False



import json
from pathlib import Path

from vosk import KaldiRecognizer
from vosk import Model as VModel

from ..config import INPUT_DEVICE_CONFIG, get_config
from ..model_manager import download_model


class Model(VModel):
    def __init__(self, model_name, lang=None):
        models_root = Path("~/.cache/voice_commands/models/").expanduser()
        model_name_path = models_root / model_name

        if not model_name_path.exists():
            print(f"Model of name <{model_name}> does not exist.")
            url = input("Enter the url for the model you want to download: ")
            download_model(url, model_name)

        model_name_path_children = [d for d in model_name_path.iterdir() if d.is_dir()]
        model_path = model_name_path / model_name_path_children[0]

        super().__init__(model_path.as_posix(), model_name, lang)
        config = get_config(INPUT_DEVICE_CONFIG)
        self.recognizer = KaldiRecognizer(self, config["rate"])

    def transcribe(self, segment):
        self.recognizer.AcceptWaveform(segment)
        text = json.loads(self.recognizer.Result())["text"]
        return text



from ..types import Command, command_callback


def handle_feedback(generator, assistant):
    for feedback in generator:
        t = type(feedback)
        if t is str:
            print(feedback)
        elif callable(t):
            handle_feedback(feedback(assistant.next_transcription()), assistant)
            break
        else:
            raise RuntimeError(
                f"got unexpected feedback type. allowed types are str | {command_callback}"
            )


def execute_command(command: Command, transcription: str, assistant):
    handle_feedback(command.callback(transcription), assistant)



import json
import select
import sys
import time
from pathlib import Path

from .core.audio_api import api
from .core.audio_processing_utils import mean_spectogram_from_buffer

CONFIG_ROOT = Path("~/.config/voice_commands").expanduser()
CONFIG_ROOT.mkdir(parents=True, exist_ok=True)

INPUT_DEVICE_CONFIG = CONFIG_ROOT / "input_device.json"
VAD_CONFIG = CONFIG_ROOT / "voice_activity_detection.json"


def set_config(config: dict, path):
    with open(path, "w") as file:
        json.dump(config, file)


def get_config(path):
    with open(path, "r") as file:
        return json.load(file)


def get_input_device_info(index: int, dtype: int):
    info = api.get_device_info_by_index(index)
    return {
        "format": dtype,
        "rate": int(info["defaultSampleRate"]),
        "input": True,
        "channels": info["maxInputChannels"],
        "input_device_index": index,
    }


def get_vad_info(input_device_info: dict):
    print(
        "This is the calibration programm. "
        "It will determine the noise and speech levels for your device."
    )
    input("Press Enter to continue...")
    print("Wait while the noise level is being determined.\n")
    mean_spectogram = []

    with api.open(**input_device_info) as stream:
        start = time.time()
        while stream.is_active():
            buffer = stream.read(512, exception_on_overflow=False)
            mean_spectogram.append(mean_spectogram_from_buffer(buffer))
            if time.time() - start > 1:
                stream.stop_stream()

    noise_level = max(mean_spectogram)
    mean_spectogram.clear()

    print("The programm will now determin the speech level.")
    print("Press Enter after you said something...")
    with api.open(**input_device_info) as stream:
        while stream.is_active():
            buffer = stream.read(512, exception_on_overflow=False)
            mean_spectogram.append(mean_spectogram_from_buffer(buffer))

            if select.select([sys.stdin], [], [], 0.0)[0]:
                input_data = sys.stdin.readline()
                if input_data:
                    break

    speech_level = max(mean_spectogram)

    return {"noise_level": noise_level, "speech_level": speech_level}


def make_config(
    device_query: str, min_speech_volume_ratio: float = 0.2, input_device_dtype: int = 8
):
    input_device = api.get_input_device_from_query(device_query)
    input_device_info = get_input_device_info(input_device["index"], input_device_dtype)
    set_config(input_device_info, INPUT_DEVICE_CONFIG)

    try:
        vad_config = get_config(VAD_CONFIG)
        if not ("speech_level" in vad_config and "noise_level" in vad_config):
            vad_config = get_vad_info(input_device_info)
        if not "min_speech_volume_ratio" in vad_config:
            vad_config["min_speech_volume_ratio"] = min_speech_volume_ratio
        set_config(vad_config, VAD_CONFIG)
    except FileNotFoundError:
        vad_info = get_vad_info(input_device_info)
        vad_info["min_speech_volume_ratio"] = min_speech_volume_ratio
        set_config(vad_info, VAD_CONFIG)
